@@ -1,10 +1,12 @@
 ;;; buttercup.el --- Behavior-Driven Emacs Lisp Testing -*-lexical-binding:t-*-
 
 ;; Copyright (C) 2015-2017  Jorgen Schaefer <contact@jorgenschaefer.de>
+;; Copyright (C) 2018-2024  Ola Nilsson <ola.nilsson@gmail.com>
 
-;; Version: 1.24
+;; Version: 1.35
 ;; Author: Jorgen Schaefer <contact@jorgenschaefer.de>
-;; Package-Requires: ((emacs "24.3"))
+;; Maintainer: Ola Nilsson <ola.nilsson@gmail.com>
+;; Package-Requires: ((emacs "24.4"))
 ;; URL: https://github.com/jorgenschaefer/emacs-buttercup
 
 ;; This program is free software; you can redistribute it and/or
@@ -30,6 +32,7 @@
 ;; the first parameter describing the suite and the rest being the body
 ;; of code that implements the suite.
 
+;; ;;; lexical binding is required -*- lexical-binding: t; -*-
 ;; (describe "A suite"
 ;;   (it "contains a spec with an expectation"
 ;;     (expect t :to-be t)))
@@ -47,40 +50,74 @@
 (require 'ert nil t)
 (require 'warnings)
 
+;; A base error for all errors raised by buttercup.
+(define-error 'buttercup-error-base "error")
+;; Buttercup internals error, raised on internal implementation
+;; inconsistencies.
+(define-error 'buttercup-internals-error "Internal buttercup error"
+              'buttercup-error-base)
+;; Raised when expanding `describe` macros whithout lexical-binding: t
+(define-error 'buttercup-dynamic-binding-error "Lexical binding is not enabled"
+              'buttercup-error-base)
+
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 ;;; wrapper function manipulation
 
+;; Error for buttercup--enclosed-expr
+(define-error 'buttercup-enclosed-expression-error
+              "Bad test expression"
+              'buttercup-internals-error)
+
+(eval-and-compile
+  (when (fboundp 'oclosure-define)      ;Emacs≥29
+    (oclosure-define (buttercup--thunk (:predicate buttercup--thunk-p))
+      "An elisp expression as a function and original code."
+      expr)))
+
 (defun buttercup--enclosed-expr (fun)
-  "Given a zero-arg function FUN, return its unevaluated expression.
+  "Given a FUN `buttercup-thunk', return its unevaluated expression.
 
-The function MUST have one of the following forms:
+For Emacs < 29:
+The function MUST be byte-compiled or have one of the following
+forms:
 
-\(lambda () EXPR)
-\(lambda () (buttercup--mark-stackframe) EXPR)
-\(closure (ENVLIST) () EXPR)
-\(closure (ENVLIST) () (buttercup--mark-stackframe) EXPR)
-\(lambda () (quote EXPR) EXPR)
-\(closure (ENVLIST) () (quote EXPR) EXPR)
+\(closure (ENVLIST) () (quote EXPR) (buttercup--mark-stackframe) EXPANDED)
+\(lambda () (quote EXPR) (buttercup--mark-stackframe) EXPR)
 
-and the return value will be EXPR, unevaluated. The latter 2
-forms are useful if EXPR is a macro call, in which case the
-`quote' ensures access to the un-expanded form."
+and the return value will be EXPR, unevaluated. The quoted EXPR
+is useful if EXPR is a macro call, in which case the `quote'
+ensures access to the un-expanded form."
+  (cl-assert (functionp fun) t "Expected FUN to be a function")
+  (if (and (fboundp 'buttercup--thunk-p)      ;Emacs≥29
+           (buttercup--thunk-p fun))
+      (buttercup--thunk--expr fun)
   (pcase fun
-    (`(closure ,(pred listp) nil ,expr) expr)
-    (`(closure ,(pred listp) nil (buttercup--mark-stackframe) ,expr) expr)
-    (`(closure ,(pred listp) nil (quote ,expr) . ,_rest) expr)
-    (`(closure ,(pred listp) nil ,_expr . ,(pred identity))
-     (error "Closure contains multiple expressions: %S" fun))
-    (`(closure ,(pred listp) ,(pred identity) . ,(pred identity))
-     (error "Closure has nonempty arglist: %S" fun))
-    (`(lambda nil ,expr) expr)
-    (`(lambda nil (buttercup--mark-stackframe) ,expr) expr)
-    (`(lambda nil (quote ,expr) . ,_rest) expr)
-    (`(lambda nil ,_expr . ,(pred identity))
-     (error "Function contains multiple expressions: %S" fun))
-    (`(lambda ,(pred identity) . ,(pred identity))
-     (error "Function has nonempty arglist: %S" fun))
-    (_ (error "Not a zero-arg one-expression closure: %S" fun))))
+    ;; This should be the normal case, a closure with unknown enclosed
+    ;; variables, empty arglist and a body containing
+    ;; * the quoted original expression
+    ;; * the stackframe marker
+    ;; * the macroexpanded original expression
+    (`(closure ,(pred listp) nil
+        (quote ,expr) (buttercup--mark-stackframe) ,_expanded)
+     expr)
+    ;; This a when FUN has not been evaluated.
+    ;; Why does that happen?
+    ;; A lambda with an empty arglist and a body containing
+    ;; * the quoted original expression
+    ;; * the stackframe marker
+    ;; * the expanded expression
+    (`(lambda nil
+        (quote ,expr) (buttercup--mark-stackframe) ,_expanded)
+     expr)
+    ;; This is when FUN has been byte compiled, as when the entire
+    ;; test file has been byte compiled. Check that it has an empty
+    ;; arglist, that is all that is possible at this point. The
+    ;; return value is byte compiled code, not the original
+    ;; expressions. Also what is possible at this point.
+    ((and (pred byte-code-function-p) (guard (member (aref fun 0) '(nil 0))))
+     (aref fun 1))
+    ;; Error
+    (_ (signal 'buttercup-enclosed-expression-error (format "Not a zero-arg one-expression closure: %S" fun))))))
 
 (defun buttercup--expr-and-value (fun)
   "Given a function, return its quoted expression and value.
@@ -112,14 +149,51 @@ a call to `save-match-data', as `format-spec' modifies that."
   (save-match-data
     (format-spec format specification)))
 
+(defun buttercup--simple-format (specification &rest format)
+  "Return a string based on SPECIFICATION and FORMAT.
+A simpler version of `format-spec', which see.
+If more than one FORMAT string is given they will be combined
+before formatting replacements occur.
+
+Does not support flags, width or precision.
+The substitution for a specification character can be a function,
+which is only supported in `format-spec' from Emacs 29.
+Does not have the IGNORE-MISSING and SPLIT parameters."
+  (save-match-data
+    (with-temp-buffer
+      (apply #'insert format)
+      (goto-char 1)
+      (while (search-forward "%" nil t)
+        (cond
+         ((= (following-char) ?%)
+          (delete-char 1))
+         ((looking-at-p (rx alpha))
+          (let* ((char (following-char))
+                 (begin (point))
+                 (replacement (cdr (assq char specification)))
+                 (text (if (functionp replacement) (funcall replacement) replacement)))
+            (insert-and-inherit text)
+            (delete-char 1)
+            (delete-region (1- begin) begin)))))
+      (buffer-string))))
+
 ;;;;;;;;;;
 ;;; expect
 
-(define-error 'buttercup-failed
-  "Buttercup test failed")
+(define-error 'buttercup-failed "Buttercup test failed" 'buttercup-error-base)
 
-(define-error 'buttercup-pending
-  "Buttercup test is pending")
+(define-error 'buttercup-pending "Buttercup test is pending" 'buttercup-error-base)
+
+(defun buttercup--wrap-expr (expr)
+  "Wrap EXPR in a `buttercup--thunk' to be used by `buttercup-expect'."
+  (if (fboundp 'oclosure-lambda)        ;Emacs≥29
+      `(oclosure-lambda (buttercup--thunk (expr ',expr)) ()
+         (buttercup--mark-stackframe)
+         ,expr)
+    `(lambda ()
+       (quote ,expr)
+       (buttercup--mark-stackframe)
+       ,expr)))
 
 (defmacro expect (arg &optional matcher &rest args)
   "Expect a condition to be true.
@@ -135,19 +209,9 @@ This macro knows three forms:
 
 \(expect ARG)
   Fail the current test if ARG is not true."
-  (let ((wrapped-args
-         (mapcar (lambda (expr) `(lambda ()
-                                   (quote ,expr)
-                                   (buttercup--mark-stackframe)
-                                   ,expr))
-                 args)))
-    `(buttercup-expect
-      (lambda ()
-        (quote ,arg)
-        (buttercup--mark-stackframe)
-        ,arg)
-      ,(or matcher :to-be-truthy)
-      ,@wrapped-args)))
+  `(buttercup-expect ,(buttercup--wrap-expr arg)
+                     ,(or matcher :to-be-truthy)
+                     ,@(mapcar #'buttercup--wrap-expr args)))
 
 (defun buttercup-expect (arg &optional matcher &rest args)
   "The function for the `expect' macro.
@@ -209,7 +273,8 @@ The BODY will receive ARGS as functions that can be called (using
 simple boolean, or a cons cell of the form (RESULT . MESSAGE). If
 RESULT is nil, MESSAGE should describe why the matcher failed. If
 RESULT is non-nil, MESSAGE should describe why a negated matcher
-failed."
+failed.
+BODY may start with a docstring."
   (declare (indent defun))
   `(put ,matcher 'buttercup-matcher
         (lambda ,args
@@ -333,7 +398,7 @@ See also `buttercup-define-matcher'."
                  (concat expect-match-phrase " "
                          expect-mismatch-phrase))))
       (if (symbolp function)
-          (setq function-name function)
+          (setq function-name (symbol-name function))
         (error "The `:function-name' keyword is required if FUNCTION is not a symbol")))
     `(buttercup-define-matcher ,matcher (arg)
        (let* ((expr (buttercup--enclosed-expr arg))
@@ -409,7 +474,7 @@ See also `buttercup-define-matcher'."
                  (concat expect-match-phrase " "
                          expect-mismatch-phrase))))
       (if (symbolp function)
-          (setq function-name function)
+          (setq function-name (symbol-name function))
         (error "The `:function-name' keyword is required if FUNCTION is not a symbol")))
     `(buttercup-define-matcher ,matcher (a b)
        (cl-destructuring-bind
@@ -481,7 +546,7 @@ See also `buttercup-define-matcher'."
       ((text-expr . text) (regexp-expr . regexp))
       (mapcar #'buttercup--expr-and-value (list text regexp))
     (save-match-data
-      (let* (;; For string literals, juse use them normally, but for
+      (let* (;; For string literals, just use them normally, but for
              ;; expressions, show both the expr and its string value
              (text-is-literal (equal text-expr text))
              (regexp-is-literal (equal regexp-expr regexp))
@@ -566,13 +631,19 @@ See also `buttercup-define-matcher'."
                 a-expr b tolerance a (abs (- a b)))))))
 
 (buttercup-define-matcher :to-throw (expr &optional signal signal-args)
-  (let ((expected-signal-symbol (or (and signal (funcall signal)) 'error))
+  "Check that EXPR raises SIGNAL with SIGNAL-ARGS.
+EXPR, SIGNAL, and SIGNAL-ARGS should all be buttercup-wrapped
+objects. EXPR is the test code that will be evaluated. The signal
+raised by EXPR must be SIGNAL or an error signal derived from
+SIGNAL. If SIGNAL is nil match any error signal. If SIGNAL-ARGS
+are given they must be `equal' to the arguments of the caught
+signal. Do not consider the signal arguments if SIGNAL-ARGS is
+nil."
+  (let ((expected-signal-symbol (and signal (funcall signal)))
         (expected-signal-args (and signal-args (funcall signal-args)))
         (unevaluated-expr (buttercup--enclosed-expr expr))
         expr-value
-        thrown-signal
-        thrown-signal-symbol
-        thrown-signal-args)
+        thrown-signal)
     (when (and (functionp unevaluated-expr)
                (member (car unevaluated-expr) '(lambda closure)))
       (display-warning
@@ -581,65 +652,102 @@ See also `buttercup-define-matcher'."
         (format "Probable incorrect use of `:to-throw' matcher: pass an expression instead of a function: `%S'"
                 unevaluated-expr)
         'yellow)))
-    ;; If no signal specificaiton, use `error' as the signal symbol
-    (when (and (null expected-signal-symbol)
-               (null expected-signal-args))
-      (setq expected-signal-symbol 'error))
-    ;; Set the above 4 variables
+    ;; Set the above variables
     (condition-case err
         (setq expr-value
               (funcall expr))
       (error
-       (setq thrown-signal err
-             thrown-signal-symbol (car err)
-             thrown-signal-args (cdr err))
+       (setq thrown-signal err)
        nil))
+    (buttercup--handle-to-throw thrown-signal
+                                (cons expected-signal-symbol expected-signal-args)
+                                unevaluated-expr expr-value)))
+
+(defun buttercup--handle-to-throw (thrown-signal expected-signal unevaluated-expr expr-value)
+  "Handle the results of the :to-throw matcher.
+This is a separate function for testability purposes.
+THROWN-SIGNAL is the signal - a `cons' of symbol and arguments -
+caught by `condition-case', or nil if no signal was raised.
+EXPECTED-SIGNAL is a `cons' of the expected signal symbol and
+arguments. The `cdr' can be nil if the `expect' statement did not
+specify any expected arguments.
+UNEVALUATED-EXPR is the Lisp sexp used before the :to-throw
+matcher keyword in the `expect' statement.
+EXPR-VALUE is the return value from the evaluation of
+UNEVALUATED-EXPR if it did not raise any signal."
+  (let* ((thrown-signal-symbol (car thrown-signal))
+         (thrown-signal-args (cdr thrown-signal))
+         (expected-signal-symbol (car expected-signal))
+         (expected-signal-args (cdr expected-signal))
+         (matching-signal-symbol
+          (or (null expected-signal-symbol)
+              (memq expected-signal-symbol (get thrown-signal-symbol 'error-conditions))))
+         (explained-signal-args ; nil for matched, explained or t for mismatched
+          (when expected-signal-args
+              ;; The ert-explainer for equal does an equal internally,
+              ;; so avoid calling equal twice by calling the explainer
+              ;; directly.
+              (funcall (or (get 'equal 'ert-explainer) (lambda (a b) (not (equal a b))))
+                       thrown-signal-args expected-signal-args))))
     (let*
-        ((matched
-          (and thrown-signal
-               (or (null expected-signal-symbol)
-                   (memq expected-signal-symbol (get thrown-signal-symbol 'error-conditions)))
-               (or (null expected-signal-args)
-                   (equal thrown-signal-args expected-signal-args))))
+        ((matched (and thrown-signal matching-signal-symbol (not explained-signal-args)))
+         ;; Some of these replacement are always used, there is no
+         ;; reason not to format them immediately. But e and t are not
+         ;; always used and should be delayed. Use
+         ;; buttercup--simple-format for formatting as format-spec
+         ;; does not support functions until Emacs 29
          (spec (format-spec-make
-                ?E (format "%S" unevaluated-expr)
-                ?e (format "%S" expr-value)
-                ?t (format "%S" thrown-signal)
+                ?E (format "`%S'" unevaluated-expr)
+                ?e (lambda () (format "`%S'" expr-value))
+                ?t (lambda () (format "%S" thrown-signal))
+                ?S (lambda () (format "`%S'" thrown-signal-symbol))
+                ?A (lambda ()
+                     (if expected-signal-args
+                         (format " with args `%S'" thrown-signal-args)
+                       ""))
                 ?s (if expected-signal-symbol
                        (format "a child signal of `%S'" expected-signal-symbol)
                      "a signal")
                 ?a (if expected-signal-args
                        (format " with args `%S'" expected-signal-args)
-                     "")))
-         (result-text
-          (if thrown-signal
-              (buttercup-format-spec "it threw %t" spec)
-            (buttercup-format-spec "it evaluated successfully, returning value `%e'" spec)))
-
-         (expect-match-text
-          (concat (buttercup-format-spec "Expected `%E' to throw %s%a" spec)
-                  ", but instead "
-                  result-text))
-         (expect-mismatch-text
-          (concat (buttercup-format-spec "Expected `%E' not to throw %s%a" spec)
-                  ", but "
-                  result-text)))
-      (buttercup--test-expectation matched
-        :expect-match-phrase expect-match-text
-        :expect-mismatch-phrase expect-mismatch-text))))
+                     "")
+                ?q (lambda () (format "%S" explained-signal-args))
+                )))
+      (cond
+       (matched ;; should be the most likely result
+        `(t . ,(buttercup--simple-format
+                spec "Expected %E not to throw %s%a, but it threw %S%A")))
+       ((null thrown-signal) ; no signal raised
+        `(nil . ,(buttercup--simple-format
+                  spec "Expected %E to throw %s%a, but instead it returned %e")))
+       ((and explained-signal-args (not matching-signal-symbol)) ; neither symbol nor args matched
+        `(nil . ,(buttercup--simple-format
+                  spec
+                  "Expected %E to throw %s%a, but instead it threw %S%A")))
+       (explained-signal-args ; symbol matched
+        `(nil . ,(buttercup--simple-format
+                  spec
+                  "Expected %E to signal %s%a, but instead signalled%A which does not match because %q.")))
+       ((not matching-signal-symbol) ; args matched
+        `(nil . ,(buttercup--simple-format
+                  spec
+                  "Expected %E to throw %s%a, but instead it threw %S%A")))
+       (t (error "`buttercup--handle-to-throw' could not handle args %S %S"
+                 thrown-signal expected-signal))))))
 
 (buttercup-define-matcher :to-have-been-called (spy)
-  (setq spy (funcall spy))
-  (cl-assert (symbolp spy))
+    "Check that SPY have been called at least once."
+  (cl-assert (symbolp (funcall spy)))
   (if (spy-calls-all (funcall spy))
       t
     nil))
 
 (buttercup-define-matcher :to-have-been-called-with (spy &rest args)
+  "Check that SPY has been called at least once with arguments ARGS."
   (setq spy (funcall spy))
   (cl-assert (symbolp spy))
   (setq args (mapcar #'funcall args))
-  (let* ((calls (mapcar 'spy-context-args (spy-calls-all spy))))
+  (let* ((calls (mapcar #'spy-context-args (spy-calls-all spy))))
     (cond
      ((not calls)
       (cons nil
@@ -657,13 +765,16 @@ See also `buttercup-define-matcher'."
       t))))
 
 (buttercup-define-matcher :to-have-been-called-times (spy number)
+  "Check that SPY has been called exactly NUMBER times."
   (setq spy (funcall spy)
         number (funcall number))
   (cl-assert (symbolp spy))
   (let* ((call-count (spy-calls-count spy)))
     (cond
      ((= number call-count)
-      t)
+      (cons t
+            (format "Expected `%s' to not have been called exactly %s %s, but it was."
+                    spy number (if (= number 1) "time" "times"))))
      (t
       (cons nil
             (format "Expected `%s' to have been called %s %s, but it was called %s %s"
@@ -716,8 +827,10 @@ Return CHILD."
     (cons (buttercup-suite-or-spec-parent suite-or-spec)
           (buttercup-suite-or-spec-parents (buttercup-suite-or-spec-parent suite-or-spec)))))
 
-(define-obsolete-function-alias 'buttercup-suite-parents 'buttercup-suite-or-spec-parents "emacs-buttercup 1.10")
-(define-obsolete-function-alias 'buttercup-spec-parents 'buttercup-suite-or-spec-parents "emacs-buttercup 1.10")
+(define-obsolete-function-alias 'buttercup-suite-parents
+  #'buttercup-suite-or-spec-parents "emacs-buttercup 1.10")
+(define-obsolete-function-alias 'buttercup-spec-parents
+  #'buttercup-suite-or-spec-parents "emacs-buttercup 1.10")
 
 (defun buttercup-suites-total-specs-defined (suite-list)
   "Return the number of specs defined in all suites in SUITE-LIST."
@@ -842,6 +955,9 @@ form.")
 DESCRIPTION is a string. BODY is a sequence of instructions,
 mainly calls to `describe', `it' and `before-each'."
   (declare (indent 1) (debug (&define sexp def-body)))
+  (unless lexical-binding
+    (signal 'buttercup-dynamic-binding-error
+            "buttercup requires `lexical-binding' to be t"))
   (let ((new-body
          (cond
           ((eq (elt body 0) :var)
@@ -851,7 +967,10 @@ mainly calls to `describe', `it' and `before-each'."
            `((let* ,(elt body 1)
                ,@(cddr body))))
           (t body))))
-    `(buttercup-describe ,description (lambda () ,@new-body))))
+    (if (or (memq :var new-body)
+              (memq :var* new-body))
+      `(error "buttercup: :var(*) found in invalid position of describe form \"%s\"" ,description)
+    `(buttercup-describe ,description (lambda () ,@new-body)))))
 
 (defun buttercup-describe (description body-function)
   "Function to handle a `describe' form.
@@ -1104,23 +1223,18 @@ none of them is equivalent to `:return-value nil'.
 
 (defun spy-on (symbol &optional keyword arg)
   "Create a spy (mock) for the function SYMBOL.
+The spy will track call information that can be queried with the
+:to-have-been-called matchers.
+The default spy will always return nil, but that can be changed
+by setting KEYWORD to one of:
 
-KEYWORD can have one of the following values:
+  :and-call-through -- call the original function.
 
-  :and-call-through -- Track calls, but call the original
-      function.
+  :and-return-value -- return ARG.
 
-  :and-return-value -- Track calls, but return ARG instead of
-      calling the original function.
+  :and-call-fake -- call ARG, a function that matches SYMBOL.
 
-  :and-call-fake -- Track calls, but call ARG instead of the
-      original function.
-
-  :and-throw-error -- Signal ARG as an error instead of calling
-      the original function.
-
-  nil -- Track calls, but simply return nil instead of calling
-      the original function.
+  :and-throw-error -- Signal ARG without arguments.
 
 If the original function was a command, the generated spy will
 also be a command with the same interactive form, unless
@@ -1170,7 +1284,7 @@ responsibility to ensure ARG is a command."
             (_
              (error "Invalid `spy-on' keyword: `%S'" keyword)))))
     (unless (buttercup--spy-on-and-call-replacement symbol replacement)
-      (error "Spies can only be created in `before-each'"))))
+      (error "Spies can only be created in `before-each' or `it'"))))
 
 
 (defun buttercup--spy-on-and-call-replacement (spy fun)
@@ -1223,13 +1337,24 @@ responsibility to ensure ARG is a command."
 Should always be set to a value that is not `listp', except while
 in a `buttercup-with-cleanup' environment.")
 
+(defvar native-comp-enable-subr-trampolines)
+(defvar comp-enable-subr-trampolines)
+
 (defmacro buttercup-with-cleanup (&rest body)
   "Execute BODY, cleaning spys and the rest afterwards."
-  `(let ((buttercup--cleanup-functions nil))
-     (unwind-protect (progn ,@body)
-       (dolist (fun buttercup--cleanup-functions)
-         (ignore-errors
-           (funcall fun))))))
+  `(,@(if (fboundp 'with-suppressed-warnings)
+          '(with-suppressed-warnings ((obsolete comp-enable-subr-trampolines)))
+        '(progn))
+     (let ((buttercup--cleanup-functions nil)
+           ;; Redefining certain primitive's trampolines will cause problems,
+           ;; see https://github.com/jorgenschaefer/emacs-buttercup/issues/230 and
+           ;; https://debbugs.gnu.org/cgi/bugreport.cgi?bug=61880
+           (comp-enable-subr-trampolines nil)
+           (native-comp-enable-subr-trampolines nil))
+       (unwind-protect (progn ,@body)
+         (dolist (fun buttercup--cleanup-functions)
+           (ignore-errors
+             (funcall fun)))))))
 
 (defun buttercup--add-cleanup (function)
   "Register FUNCTION for cleanup in `buttercup-with-cleanup'."
@@ -1256,11 +1381,6 @@ in a `buttercup-with-cleanup' environment.")
            nil
            buttercup--spy-contexts))
 
-(buttercup-define-matcher :to-have-been-called (spy)
-  (if (spy-calls-all (funcall spy))
-      t
-    nil))
-
 (defun spy-calls-any (spy)
   "Return t iff SPY has been called at all, nil otherwise."
   (if (spy-calls-all spy)
@@ -1273,11 +1393,11 @@ in a `buttercup-with-cleanup' environment.")
 
 (defun spy-calls-count-returned (spy)
   "Return the number of times SPY has been called successfully so far."
-  (cl-count-if 'spy-context-return-p (spy-calls-all spy)))
+  (cl-count-if #'spy-context-return-p (spy-calls-all spy)))
 
 (defun spy-calls-count-errors (spy)
   "Return the number of times SPY has been called and thrown errors so far."
-  (cl-count-if 'spy-context-thrown-p (spy-calls-all spy)))
+  (cl-count-if #'spy-context-thrown-p (spy-calls-all spy)))
 
 (defun spy-calls-args-for (spy index)
   "Return the context of the INDEXth call to SPY."
@@ -1289,7 +1409,7 @@ in a `buttercup-with-cleanup' environment.")
 
 (defun spy-calls-all-args (spy)
   "Return the arguments for every recorded call to SPY."
-  (mapcar 'spy-context-args (spy-calls-all spy)))
+  (mapcar #'spy-context-args (spy-calls-all spy)))
 
 (defun spy-calls-most-recent (spy)
   "Return the context of the most recent call to SPY."
@@ -1341,8 +1461,11 @@ each frame, and prefixes each stack frame with lambda or M to
 indicate whether it represents a normal evaluated function call
 or a macro/special form.")
 
-(defvar buttercup-color t
-  "Whether to use colors in output.")
+(defvar buttercup-color (let ((no-color (getenv "NO_COLOR")))
+                          (or (not no-color) (string= no-color "")))
+  "Whether to use colors in output.
+Will be nil if the `NO_COLOR' environment variable is set to any value
+\(see URL https://no-color.org/).")
 
 (defconst buttercup-warning-buffer-name " *Buttercup-Warnings*"
   "Buffer name used to collect warnings issued while running a spec.
@@ -1370,9 +1493,11 @@ spec, and should be killed after running the spec.")
 
 Takes directories as command line arguments, defaulting to the
 current directory."
+  (setq backtrace-on-error-noninteractive nil)
   (let ((dirs nil)
         (patterns nil)
-        (args command-line-args-left))
+        (args command-line-args-left)
+        failed-files-suite)
     (while args
       (cond
        ((equal (car args) "--")
@@ -1408,15 +1533,44 @@ current directory."
         (push (car args) dirs)
         (setq args (cdr args)))))
     (setq command-line-args-left nil)
+    (setq failed-files-suite (make-buttercup-suite
+                              :description "File failed to load correctly:"))
     (dolist (dir (or dirs '(".")))
       (dolist (file (directory-files-recursively
                      dir "\\`test-.*\\.el\\'\\|-tests?\\.el\\'"))
         ;; Exclude any hidden directory, both immediate (^.) and nested (/.) subdirs
         (when (not (string-match "\\(^\\|/\\)\\." (file-relative-name file)))
-          (load file nil t))))
+          (buttercup--load-test-file file failed-files-suite))))
     (when patterns
       (buttercup-mark-skipped patterns t))
-    (buttercup-run)))
+    (when (buttercup-suite-children failed-files-suite)
+      ;; At least one file has failed to load, add the
+      ;; failed-files-suite to the end(?) of the suite list
+      (setq buttercup-suites (append buttercup-suites
+                                     (list failed-files-suite)))))
+    (buttercup-run))
+
+(defun buttercup--load-test-file (file failure-suite)
+  "Load FILE keeping track of failures.
+If an error is raised by `load', store the error information in a
+spec and add it to FAILURE-SUITE."
+  (cl-destructuring-bind (status description stack)
+      (buttercup--funcall #'load file nil t)
+    (when (eq status 'failed)
+      (if (memq (cl-caadr description)
+                '(buttercup-dynamic-binding-error end-of-file))
+          (setq stack nil)
+      ;; Skip all of stack until load is called
+      (while (not (eq (nth 1 (car stack)) 'load))
+        (pop stack)))
+      (buttercup-suite-add-child
+       failure-suite
+       (make-buttercup-spec
+        :description file
+        :status 'failed
+        :failure-description (error-message-string (cadr description))
+        :failure-stack stack
+        :function (lambda () (ignore)))))))
 
 (defun buttercup-mark-skipped (matcher &optional reverse)
   "Mark any spec that match MATCHER as skipped.
@@ -1482,6 +1636,7 @@ If MARKDOWN-BUFFERS is empty (nil), use the current buffer."
 (defun buttercup-run-markdown ()
   "Run all test suites defined in Markdown files passed as arguments.
 A suite must be defined within a Markdown \"lisp\" code block."
+  (setq backtrace-on-error-noninteractive nil)
   (apply #'buttercup-run-markdown-buffer (mapcar #'find-file-noselect
                                                  command-line-args-left)))
 
@@ -1507,16 +1662,18 @@ have been defined."
     (or (and noerror :no-suites)
         (error "No suites defined"))))
 
+(define-error 'buttercup-run-specs-failed "buttercup-run failed" 'buttercup-error-base)
+
 (defun buttercup--run-suites (suites &optional noerror)
   "Run a list of SUITES.
-Signal an error if any spec fail. Signal no error if NOERROR is
-non-nil. Return t if all specs pass, nil if at least one spec
-fail."
+Signal a `buttercup-run-specs-failed` error if any spec fail,
+unless NOERROR is non-nil. Return t if all specs pass, nil if at
+least one spec fail."
   (funcall buttercup-reporter 'buttercup-started suites)
   (mapc #'buttercup--run-suite suites)
   (funcall buttercup-reporter 'buttercup-done suites)
   (or (zerop (buttercup-suites-total-specs-failed suites))
-      (not (or noerror (error "")))))
+      (not (or noerror (signal 'buttercup-run-specs-failed '(""))))))
 
 (defvar buttercup--before-each nil
   "A list of functions to call before each spec.
@@ -1593,7 +1750,7 @@ Do not change the global value.")
   "Update SUITE-OR-SPEC with the result of calling FUNCTION with ARGS.
 Sets the `status', `failure-description', and `failure-stack' for
 failed and pending specs."
-  (let* ((result (apply 'buttercup--funcall function args))
+  (let* ((result (apply #'buttercup--funcall function args))
          (status (elt result 0))
          (description (elt result 1))
          (stack (elt result 2)))
@@ -1663,6 +1820,15 @@ Two special statuses can be listed in
            (string= (buttercup-spec-failure-description spec) "PENDING"))
       ))
 
+(defun buttercup--reporter-batch-preprint-spec-p (spec)
+  "Return non-nil if the SPEC description should be printed at `spec-started'."
+  (not (or buttercup-reporter-batch-quiet-statuses
+           ;; Do not 'pre-print' in github actions unless color is
+           ;; disabled. See #181.
+           (and (getenv "GITHUB_ACTION") buttercup-color)
+           (and buttercup-color
+                (string-match-p "[\n\v\f]" (buttercup-spec-description spec))))))
+
 (defun buttercup-reporter-batch (event arg)
   "A reporter that handles batch sessions.
 
@@ -1687,10 +1853,8 @@ EVENT and ARG are described in `buttercup-reporter'."
            (push arg buttercup-reporter-batch--suite-stack)
          (buttercup--print "%s\n" (buttercup--indented-description arg))))
       (`spec-started
-       (or buttercup-reporter-batch-quiet-statuses
-           (and buttercup-color
-                (string-match-p "[\n\v\f]" (buttercup-spec-description arg)))
-           (buttercup--print "%s" (buttercup--indented-description arg))))
+       (when (buttercup--reporter-batch-preprint-spec-p arg)
+         (buttercup--print "%s" (buttercup--indented-description arg))))
       (`spec-done
        (when (and buttercup-reporter-batch-quiet-statuses
                   (not (buttercup-reporter-batch--quiet-spec-p arg)))
@@ -1726,8 +1890,8 @@ EVENT and ARG are described in `buttercup-reporter'."
 (defun buttercup-reporter-batch--print-spec-done-line (spec color)
   "Print the remainder of the SPEC report line for `spec-done'.
 
-If COLOR is non-nil, erace the text so far on the current line
-using '\r' and replace it with the same text colored according to
+If COLOR is non-nil, erase the text so far on the current line
+using '\\r' and replace it with the same text colored according to
 the SPEC status. Do not erase and replace if the text would have
 been reprinted with the default color.
 
@@ -1761,7 +1925,8 @@ Finally print the elapsed time for SPEC."
       ;; Carriage returns (\r) should not be colorized. It would mess
       ;; up color handling in Emacs compilation buffers using
       ;; `ansi-color-apply-on-region' in `compilation-filter-hook'.
-      (buttercup--print "\r%s"
+      (buttercup--print "%s%s"
+                        (if (buttercup--reporter-batch-preprint-spec-p spec) "\r" "")
                         (buttercup-colorize
                          (buttercup--indented-description spec) color)))
     (unless (eq 'passed status)
@@ -1827,8 +1992,7 @@ Colorize parts of the output if COLOR is non-nil."
 FMT and ARGS are passed to `format'."
   (send-string-to-terminal (apply #'format fmt args)))
 
-
-(defadvice display-warning (around buttercup-defer-warnings activate)
+(defun buttercup--display-warning (fn type message &optional level buffer-name &rest args)
   "Log all warnings to a special buffer while running buttercup specs.
 
 Emacs' normal display logic for warnings doesn't mix well with
@@ -1848,8 +2012,10 @@ finishes."
       (cl-letf
           ((warning-minimum-level :emergency)
            ((symbol-function 'message) 'ignore))
-        ad-do-it)
-    ad-do-it))
+        (apply fn type message level buffer-name args))
+    (apply fn type message level buffer-name args)))
+
+(advice-add 'display-warning :around #'buttercup--display-warning)
 
 (defconst buttercup-colors
   '((black   . 30)
@@ -1877,8 +2043,8 @@ the capturing behavior."
 (defun buttercup-colorize (string color)
   "Format STRING with COLOR.
 Return STRING unmodified if COLOR is nil."
-  (if color
-      (let ((color-code (cdr (assoc color buttercup-colors))))
+  (if (and color buttercup-color)
+      (let ((color-code (cdr (assq color buttercup-colors))))
         (format "\e[%sm%s\e[0m" color-code string))
     string))
 
@@ -1899,9 +2065,10 @@ EVENT and ARG are described in `buttercup-reporter'."
                                 (with-current-buffer buf
                                   (let ((inhibit-read-only t))
                                     (goto-char (point-max))
-                                    (insert (apply 'format fmt args))))))
+                                    (insert (apply #'format fmt args))))))
       (unwind-protect
-          (buttercup-reporter-batch event arg)
+          (let ((buttercup-color))
+            (buttercup-reporter-batch event arg))
         (fset 'buttercup--print old-print)))
     (let ((w (get-buffer-window (current-buffer))))
       (when w
@@ -1931,15 +2098,19 @@ failed -- The second value is the description of the expectation
 
 (defun buttercup--debugger (&rest args)
   "Debugger function that return error context with an exception.
-
 ARGS according to `debugger'."
   ;; If we do not do this, Emacs will not run this handler on
   ;; subsequent calls. Thanks to ert for this.
   (setq num-nonmacro-input-events (1+ num-nonmacro-input-events))
   (throw 'buttercup-debugger-continue
-         (list 'failed args (buttercup--backtrace))))
+         (list 'failed args
+               ;; args is (error (signal . data) ....) where the tail
+               ;; may be empty
+               (cl-destructuring-bind (signal-type . data) (cl-second args)
+                 (unless (eq signal-type 'buttercup-pending)
+                   (buttercup--backtrace))))))
 
-(defalias 'buttercup--mark-stackframe 'ignore
+(defalias 'buttercup--mark-stackframe #'ignore
   "Marker to find where the backtrace start.")
 
 (defun buttercup--backtrace ()
@@ -2039,6 +2210,7 @@ With buttercup minor mode active the following is activated:
 - `describe' and `it' forms are available from `imenu' for
   quicker access."
   :lighter " ❀"
+  :keymap (make-sparse-keymap)
   (let ((font-lock-form '(("(\\(describe\\|buttercup-define-matcher\\|it\\) "
                            1 'font-lock-keyword-face)))
         (imenu-forms '(("Test Suites" "\\((describe\\_> +\\)\"\\(\\_<.+\\_>\\)\"" 2)
@@ -2078,9 +2250,10 @@ ARG is ‘toggle’; disable otherwise."
            (add-hook 'after-load-functions #'buttercup-check-for-stale-elc)))
         (t (remove-hook 'after-load-functions #'buttercup-check-for-stale-elc))))
 
+(provide 'buttercup)
 ;; Local Variables:
 ;; indent-tabs-mode: nil
 ;; sentence-end-double-space: nil
+;; tab-width: 8
 ;; End:
-(provide 'buttercup)
 ;;; buttercup.el ends here
